@@ -5,7 +5,7 @@ from typing import Optional
 import pandas as pd
 import numpy as np
 import pulp as plp
-from pulp import PULP_CBC_CMD
+from pulp import PULP_CBC_CMD, GLPK_CMD
 import logging
 
 class optimization:
@@ -16,8 +16,6 @@ class optimization:
     
     This class methods are:
 
-    - get_load_unit_cost
-
     - perform_optimization
 
     - perform_perfect_forecast_optim
@@ -28,7 +26,8 @@ class optimization:
 
     def __init__(self, retrieve_hass_conf: dict, optim_conf: dict, plant_conf: dict, 
                  var_load_cost: str, var_prod_price: str, days_list: pd.date_range, 
-                 costfun: str, config_path: str, logger: logging.Logger, opt_time_delta: Optional[int] = 24) -> None:
+                 costfun: str, config_path: str, logger: logging.Logger, 
+                 opt_time_delta: Optional[int] = 24) -> None:
         """
         Define constructor for optimization class.
         
@@ -160,12 +159,21 @@ class optimization:
         for i in set_I:
             P_def_sum.append(plp.lpSum(P_deferrable[k][i] for k in range(self.optim_conf['num_def_loads'])))
         if self.costfun == 'profit':
-            objective = plp.lpSum(-0.001*self.timeStep*(unit_load_cost[i]*(P_load[i] + P_def_sum[i]) + \
-                                                        unit_prod_price[i] * P_grid_neg[i])
-                                  for i in set_I)
+            if self.optim_conf['set_total_pv_sell']:
+                objective = plp.lpSum(-0.001*self.timeStep*(unit_load_cost[i]*(P_load[i] + P_def_sum[i]) + \
+                                                            unit_prod_price[i]*P_grid_neg[i])
+                                      for i in set_I)
+            else:
+                objective = plp.lpSum(-0.001*self.timeStep*(unit_load_cost[i]*P_grid_pos[i] + \
+                                                            unit_prod_price[i]*P_grid_neg[i])
+                                      for i in set_I)
         elif self.costfun == 'cost':
-            objective = plp.lpSum(-0.001*self.timeStep*unit_load_cost[i]*(P_load[i] + P_def_sum[i])
-                                  for i in set_I)
+            if self.optim_conf['set_total_pv_sell']:
+                objective = plp.lpSum(-0.001*self.timeStep*unit_load_cost[i]*P_grid_pos[i]
+                                      for i in set_I)
+            else:
+                objective = plp.lpSum(-0.001*self.timeStep*unit_load_cost[i]*(P_load[i] + P_def_sum[i])
+                                      for i in set_I)
         elif self.costfun == 'self-consumption':
             objective = plp.lpSum(0.001*self.timeStep*unit_load_cost[i]*SC[i] for i in set_I)
         else:
@@ -214,7 +222,7 @@ class optimization:
         for k in range(self.optim_conf['num_def_loads']):
             constraints.update({"constraint_defload{}_energy".format(k) :
                                 plp.LpConstraint(
-                                    e = plp.lpSum(P_deferrable[k][i] for i in set_I),
+                                    e = plp.lpSum(P_deferrable[k][i]*self.timeStep for i in set_I),
                                     sense = plp.LpConstraintEQ,
                                     rhs = self.optim_conf['def_total_hours'][k]*self.optim_conf['P_deferrable_nom'][k])
                                 })
@@ -299,7 +307,18 @@ class optimization:
     
         ## Finally, we call the solver to solve our optimization model:
         # solving with default solver CBC
-        opt_model.solve(PULP_CBC_CMD(msg=0))
+        try:
+            opt_model.solve(PULP_CBC_CMD(msg=0))
+        except Exception:
+            self.logger.warning("Failed LP solve with PULP_CBC_CMD solver, falling back to default Pulp")
+            try:
+                opt_model.solve()
+            except Exception:
+                self.logger.warning("Failed LP solve with default Pulp solver, falling back to GLPK_CMD")
+                try:
+                    opt_model.solve(GLPK_CMD(msg=0))
+                except Exception:
+                    self.logger.error("It was not possible to find a valid solver for Pulp package")
         
         # The status of the solution is printed to the screen
         self.logger.info("Status: " + plp.LpStatus[opt_model.status])
@@ -333,21 +352,26 @@ class optimization:
             P_def_sum_tp.append(sum(P_deferrable[k][i].varValue for k in range(self.optim_conf['num_def_loads'])))
         opt_tp["unit_load_cost"] = [unit_load_cost[i] for i in set_I]
         opt_tp["unit_prod_price"] = [unit_prod_price[i] for i in set_I]
-        opt_tp["cost_profit"] = [-0.001*self.timeStep*(unit_load_cost[i]*(P_load[i] + P_def_sum_tp[i]) + \
-                                                       unit_prod_price[i] * P_grid_neg[i].varValue)
-                                 for i in set_I]
+        if self.optim_conf['set_total_pv_sell']:
+            opt_tp["cost_profit"] = [-0.001*self.timeStep*(unit_load_cost[i]*(P_load[i] + P_def_sum_tp[i]) + \
+                unit_prod_price[i]*P_grid_neg[i].varValue) for i in set_I]
+        else:
+            opt_tp["cost_profit"] = [-0.001*self.timeStep*(unit_load_cost[i]*P_grid_pos[i].varValue + \
+                unit_prod_price[i]*P_grid_neg[i].varValue) for i in set_I]
         
         if self.costfun == 'profit':
-            opt_tp["cost_fun_profit"] = [-0.001*self.timeStep*(unit_load_cost[i]*(P_load[i] + P_def_sum_tp[i]) + \
-                                                               unit_prod_price[i] * P_grid_neg[i].varValue)
-                                         for i in set_I]
+            if self.optim_conf['set_total_pv_sell']:
+                opt_tp["cost_fun_profit"] = [-0.001*self.timeStep*(unit_load_cost[i]*(P_load[i] + P_def_sum_tp[i]) + \
+                    unit_prod_price[i]*P_grid_neg[i].varValue) for i in set_I]
+            else:
+                opt_tp["cost_fun_profit"] = [-0.001*self.timeStep*(unit_load_cost[i]*P_grid_pos[i].varValue + \
+                    unit_prod_price[i]*P_grid_neg[i].varValue) for i in set_I]
         elif self.costfun == 'cost':
-            opt_tp["cost_fun_cost"] = [-0.001*self.timeStep*unit_load_cost[i]*(P_load[i] + P_def_sum_tp[i])
-                                       for i in set_I]
+            opt_tp["cost_fun_cost"] = [-0.001*self.timeStep*unit_load_cost[i]*(P_load[i] + P_def_sum_tp[i]) for i in set_I]
         elif self.costfun == 'self-consumption':
             opt_tp["cost_fun_selfcons"] = [-0.001*self.timeStep*unit_load_cost[i]*SC[i].varValue for i in set_I]
         else:
-            self.logger.error("The cost function specified type is not valid") 
+            self.logger.error("The cost function specified type is not valid")
         
         return opt_tp
 
