@@ -6,6 +6,7 @@ import numpy as np, pandas as pd
 from requests import get, post
 import json
 import datetime, logging
+from emhass.utils import set_df_index_freq
 
 class retrieve_hass:
     """
@@ -25,7 +26,8 @@ class retrieve_hass:
     """
 
     def __init__(self, hass_url: str, long_lived_token: str, freq: pd.Timedelta, 
-                 time_zone: datetime.timezone, config_path: str, logger: logging.Logger) -> None:
+                 time_zone: datetime.timezone, params: str, config_path: str, logger: logging.Logger,
+                 get_data_from_file: Optional[bool] = False) -> None:
         """
         Define constructor for retrieve_hass class.
         
@@ -37,17 +39,26 @@ class retrieve_hass:
         :type freq: pd.TimeDelta
         :param time_zone: The time zone
         :type time_zone: datetime.timezone
+        :param params: Configuration parameters passed from data/options.json
+        :type params: str
         :param config_path: The path to the yaml configuration file
         :type config_path: str
         :param logger: The passed logger object
         :type logger: logging object
+        :param get_data_from_file: Select if data should be retrieved from a 
+        previously saved pickle useful for testing or directly from connection to
+        hass database
+        :type get_data_from_file: bool, optional
 
         """
         self.hass_url = hass_url
         self.long_lived_token = long_lived_token
         self.freq = freq
         self.time_zone = time_zone
+        self.params = params
+        self.config_path = config_path
         self.logger = logger
+        self.get_data_from_file = get_data_from_file
 
     def get_data(self, days_list: pd.date_range, var_list: list, minimal_response: Optional[bool] = False,
                  significant_changes_only: Optional[bool] = False) -> None:
@@ -75,42 +86,53 @@ class retrieve_hass:
         """
         self.logger.info("Retrieve hass get data method initiated...")
         self.df_final = pd.DataFrame()
+        # Looping on each day from days list
         for day in days_list:
         
             for i, var in enumerate(var_list):
                 
-                url = self.hass_url+"api/history/period/"+day.isoformat()+"?filter_entity_id="+var
-                if minimal_response:
+                if self.params is not None: # If this is the case we suppose that we are using the supervisor API
+                    url = self.hass_url+"/history/period/"+day.isoformat()+"?filter_entity_id="+var
+                else: # Otherwise the Home Assistant Core API it is
+                    url = self.hass_url+"api/history/period/"+day.isoformat()+"?filter_entity_id="+var
+                if minimal_response: # A support for minimal response
                     url = url + "?minimal_response"
-                if significant_changes_only:
+                if significant_changes_only: # And for signicant changes only (check the HASS restful API for more info)
                     url = url + "?significant_changes_only"
                 headers = {
                     "Authorization": "Bearer " + self.long_lived_token,
                     "content-type": "application/json",
                 }
                 response = get(url, headers=headers)
-                try:
+                try: # Sometimes when there are connection problems we need to catch empty retrieved json
                     data = response.json()[0]
                 except IndexError:
                     self.logger.error("The retrieved JSON is empty, check that correct day or variable names are passed")
+                    self.logger.error("Either the names of the passed variables are not correct or days_to_retrieve is larger than the recorded history of your sensor (check your recorder settings)")
                     break
                 df_raw = pd.DataFrame.from_dict(data)
                 if len(df_raw) == 0:
                     self.logger.error("Retrieved empty Dataframe, check that correct day or variable names are passed")
-                if i == 0:
+                    self.logger.error("Either the names of the passed variables are not correct or days_to_retrieve is larger than the recorded history of your sensor (check your recorder settings)")
+                if i == 0: # Defining the DataFrame container
                     from_date = pd.to_datetime(df_raw['last_changed']).min()
                     to_date = pd.to_datetime(df_raw['last_changed']).max()
                     ts = pd.to_datetime(pd.date_range(start=from_date, end=to_date, freq=self.freq), 
                                         format='%Y-%d-%m %H:%M').round(self.freq)
                     df_day = pd.DataFrame(index = ts)
+                # Caution with undefined string data: unknown, unavailable, etc.
                 df_tp = df_raw.copy()[['state']].replace(
                     ['unknown', 'unavailable', ''], np.nan).astype(float).rename(columns={'state': var})
+                # Setting index, resampling and concatenation
                 df_tp.set_index(pd.to_datetime(df_raw['last_changed']), inplace=True)
                 df_tp = df_tp.resample(self.freq).mean()
                 df_day = pd.concat([df_day, df_tp], axis=1)
             
             self.df_final = pd.concat([self.df_final, df_day], axis=0)
-        self.df_final.index.freq = self.freq
+        self.df_final = set_df_index_freq(self.df_final)
+        if self.df_final.index.freq != self.freq:
+            self.logger.error("The inferred freq from data is not equal to the defined freq in passed parameters")
+        #self.df_final.index.freq = self.freq
 
     
     def prepare_data(self, var_load: str, load_negative: Optional[bool] = False, set_zero_min: Optional[bool] = True,
@@ -137,16 +159,17 @@ class retrieve_hass:
         :rtype: pandas.DataFrame
         
         """
-        if load_negative:
+        if load_negative: # Apply the correct sign to load power
             self.df_final[var_load+'_positive'] = -self.df_final[var_load]
         else:
             self.df_final[var_load+'_positive'] = self.df_final[var_load]
         self.df_final.drop([var_load], inplace=True, axis=1)
-        if set_zero_min:
+        if set_zero_min: # Apply minimum values
             self.df_final.clip(lower=0.0, inplace=True, axis=1)
             self.df_final.replace(to_replace=0.0, value=np.nan, inplace=True)
         new_var_replace_zero = []
         new_var_interp = []
+        # Just changing the names of variables to contain the fact that they are considered positive
         if var_replace_zero is not None:
             for string in var_replace_zero:
                 new_string = string.replace(var_load, var_load+'_positive')
@@ -159,11 +182,13 @@ class retrieve_hass:
                 new_var_interp.append(new_string)
         else:
             new_var_interp = None
+        # Treating NaN replacement: either by zeros or by linear interpolation
         if new_var_replace_zero is not None:
             self.df_final[new_var_replace_zero] = self.df_final[new_var_replace_zero].fillna(0.0)
         if new_var_interp is not None:
             self.df_final[new_var_interp] = self.df_final[new_var_interp].interpolate(
                 method='linear', axis=0, limit=None)
+        # Setting the correct time zone on DF index
         if self.time_zone is not None:
             self.df_final.index = self.df_final.index.tz_convert(self.time_zone)
         # Drop datetimeindex duplicates on final DF
@@ -188,11 +213,15 @@ class retrieve_hass:
         :type friendly_name: str
 
         """
-        url = self.hass_url+"api/states/"+entity_id
+        if self.params is not None: # If this is the case we suppose that we are using the supervisor API
+            url = self.hass_url+"/states/"+entity_id
+        else: # Otherwise the Home Assistant Core API it is
+            url = self.hass_url+"api/states/"+entity_id
         headers = {
             "Authorization": "Bearer " + self.long_lived_token,
             "content-type": "application/json",
         }
+        # Preparing the data dict to be published
         data = {
             "state": str(data_df.loc[data_df.index[idx]]),
             "attributes": {
@@ -200,7 +229,13 @@ class retrieve_hass:
                 "friendly_name": friendly_name
             }
         }
-        response = post(url, headers=headers, data=json.dumps(data))
+        # Actually post the data
+        if self.get_data_from_file:
+            class response: pass
+            response.status_code = 400
+        else:
+            response = post(url, headers=headers, data=json.dumps(data))
+        # Treating the response status and posting them on the logger
         if response.status_code == 200:
             self.logger.info("Successfully posted value in existing entity_id")
         elif response.status_code == 201:
@@ -215,4 +250,4 @@ class retrieve_hass:
             self.logger.info("Error posting value to HASS: Method not allowed")
         else:
             self.logger.info("The received response code is not recognized")
-        
\ No newline at end of file
+        return response
